//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Boresight.cc --- implementation of Boresight class
//
//	DESCRIPTION:
//		This implements all methods declared in the Boresight class
//
//	CHANGE HISTORY:
//		UL, 1999-08-17: 1.0.0: - first version from CalAtomsCommon
//		UL, 2001-03-02: 1.1.0: - added withSacCoordOrientation()
//							   - added projectOntoTelCoord()
//
//	AUTHOR:
//		U. Lammers (UL), Tue Aug 17 11:54:35 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/Boresight.cc,v 1.20 2013/03/20 10:32:45 rsaxton Exp $
//

#include <cmath>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "Boresight.h"
#include "GeometryDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( Boresight )--------------------------------------------------
//
Boresight::Boresight(const EulerAngles &bsAngles,
					 CalAtomCollection<Boresight> *parent)
	: CalChild<CalAtomCollection<Boresight> >(parent), _angles(bsAngles),
		_mIsValidForTime(-1.), _anglesIsValidForTime(-1.),
		_saccoordDefined(false)
{
	_m = _angles.toMatrix();
} // Boresight::Boresight()


const EulerAngles &Boresight::angles(double time)
{
  // Calculate the boresight angles and matrix for this time
        updateBoresight(time);

	_anglesIsValidForTime = time;

	return _angles;
} // Boresight::angles()


void Boresight::angles(const EulerAngles &eulerAngles)
{
	_angles = eulerAngles;
	_m = _angles.toMatrix();
} // Boresight::angles()


const AttitudeMatrix &Boresight::matrix(double time) 
{
  // Calculate the boresight angles and matrix for this time
        updateBoresight(time);

	_mIsValidForTime = time;

	return _m;
} // Boresight::matrix()

const AttitudeMatrix &Boresight::matrix() const
{
  // return the boresight matrix for the previously set time or for time zero
        return _m;
} // Boresight::matrix()

const Boresight &Boresight::withSacCoordOrientation(
										const SpacecraftAttitude &att) const
{
	_saccoordMatrix = att.toAttitudeMatrix();
	_saccoordDefined = true;

	return *this;
} // Boresight::withSacCoordOrientation()


XmmTelescopePointList Boresight::projectOntoTelCoord(
										const EquatorialDirection &dir) const
{
	vector<EquatorialDirection> v;
	v.push_back(dir);

	return projectOntoTelCoord(v);
} // projectOntoTelCoord()


XmmTelescopePointList Boresight::projectOntoTelCoord(
								const vector<EquatorialDirection> &dl) const
{
	if (!_saccoordDefined)
		errstr<<"Define first orientation of SACCORD"
			  <<Sas::error("SacCoordOrientationUndefined");

	// AttitudeMatrix BSA = matrix(0.) * _saccoordMatrix;
	AttitudeMatrix BSA = matrix() * _saccoordMatrix;

	unsigned N = dl.size();
	CalReal64Vector	theta(0, N), phi(0, N);
	for (unsigned n=0; n<N; ++n) {
		CartesianVector srcVec = dl[n].toUnitCartesian(),
						v = BSA * srcVec;
		theta[n] = acos(v.x())*RAD_TO_DEG*3600.,
		phi[n] = fmod(3./2.*M_PI + atan2(v.z(), v.y()), M_TWO_PI);
	}

	GeometryDataServer *gds = 0; gds = calServer.getAtom(gds);

	return gds->telCoordList(theta, phi);
} // Boresight::projectOntoTelCoord()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<Boresight>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	Boresight *null = 0;
	_collection = new vector<Boresight *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::BoreSight);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet,
							e->queryBlockName(Ccf::toString(Ccf::BoreSight)),1); // Know about algo=0 or 1
	const Table *table = ccftab.block();

      /* Check that the algoid is compatible */
        unsigned short algoid = ccftab.algoid();
        if (algoid > 1)
           errstr<<"Unknown ALGOID "<<algoid<< " in Boresight construction"
                          <<Sas::error("UnknownAlgoId");
        if (algoid < 1)
           errstr<<"Old ALGOID "<<algoid<< " in Boresight construction "<<
                   "please update your CCF" <<Sas::error("UnknownAlgoId");

	ColumnData *cdex = table->column("EULER_PHI")->data(),
			   *cdey = table->column("EULER_THETA")->data(),
			   *cdez = table->column("EULER_PSI")->data(),
			   *cddesc = table->column("INSTRUMENT_ID")->data();
	real64	*ex = cdex->real64Data(),
			*ey = cdey->real64Data(),
			*ez = cdez->real64Data();

	EulerAngles ea(0., 0., 0., EulerAngles::Euler321);
	//	look for the row corresponding to current instrument
	string instrString = Xmm::toString(instrument->id());
	bool read = false;
	for (unsigned n=0; n<table->rows(); ++n)
		if (stringColumnRow(cddesc, n) == instrString) {
			//	row corresponding to current instrument found
			ea.phi(ex[n]);
			ea.theta(ey[n]);
			ea.psi(ez[n]); 
			read = true;
			break;
		}

	if (!read)
		errstr<<"The CCF boresight misalignment file "
			  <<dataSet->qualifiedName()
			  <<" does not contain an entry for "
			  <<Xmm::toString(instrument->id())
			  <<Sas::error("NoSuchEntry");

	Boresight *&bsm = (*_collection)[0];
	if (!bsm)
		bsm = new Boresight(ea, this);

        // Read in the time-dependent boresight offsets
        bsm->readTimeDepParams(dataSet, instrument, ea);

	delete cdex;
	delete cdey;
	delete cdez;
	delete cddesc;

	dataSetServer->close(dataSet);
} // CalAtomCollection<Boresight>::CalAtomCollection()

/* 
 * Method: readTimeDepParams(dataset, instrument);
 * Purpose: To read time dependent deltas to the boresight from the CCF
 */
void Boresight::readTimeDepParams(DataSet *dataSet, const XmmInstrument *instrument, EulerAngles ea)
{

   // The deltas are stored in tables called 'INST'_ANGVAR
        string deltaTable=Xmm::toString(instrument->id()) + "_ANGVAR";

   // If the table exists then read the time dependent deltas
        if (dataSet->hasBlock(deltaTable)) {
           CcfConstituentTable ccftab(dataSet, deltaTable);
           const Table *table = ccftab.block();

           CalOalUtility::fillVector(table, "MOD_JD", _dTime);
           CalOalUtility::fillVector(table, "DELTAPHI", _dPhi);
           CalOalUtility::fillVector(table, "DELTATHETA", _dTheta);

           _timeDepParSet = true;
        } else {
           _timeDepParSet = false;
        }

   // Set the base angles
        _base_angles = ea;
}

/* 
 * Method: updateBoresight(time)
 * Purpose: modifies the base boresight with a time dependent offset
 * Input: time in seconds since start of mission
 */
void Boresight::updateBoresight(double intime)
{
    // Process boresight deltas if set, otherwise return
     if (! _timeDepParSet) return;

     //cout<<"BASE ANGLES:"<<" "<<_base_angles<<endl;

    // Convert input time into MJD
     STime t1(Xmm::missionReferenceTime);
     double mjd_diff = intime / 86400.0;  // Convert into days since mission
     double time = t1.modifiedJulianDay().full() + mjd_diff;

    // Interpolate deltas
     double deltaphi=0.0, deltatheta=0.0;
     int last = _dTime.size()-1;

     if (time < _dTime[0]) {
        deltaphi=_dPhi[0];
        deltatheta=_dTheta[0];
     } else if (time > _dTime[last]) {
        deltaphi=_dPhi[last];
        deltatheta=_dTheta[last];
     } else {

        for (int t=0; t<last-1; ++t) {

            if (time >= _dTime[t] && time <= _dTime[t+1]) 
            {
               deltaphi=_dPhi[t] + (_dPhi[t+1] - _dPhi[t]) *
                             (time - _dTime[t]) / (_dTime[t+1] - _dTime[t]);

               deltatheta=_dTheta[t] + (_dTheta[t+1] - _dTheta[t]) *
                             (time - _dTime[t]) / (_dTime[t+1] - _dTime[t]);
            }
        }
     }

        errstr << "Boresight being updated to input Time "
         << time <<" "<<intime<<" for DeltaTheta="<<deltatheta
         <<" DeltaPhi="<<deltaphi
         << Sas::message(Msg::App, Msg::Noisy);

    // Set the new angles and reset the matrix 
     EulerAngles eu = _base_angles;
     eu.theta(_base_angles.theta() +deltatheta);
     eu.phi(_base_angles.phi() +deltaphi);
     
     this->angles(eu);
} 

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
Boresight *CalAtomCollection<Boresight>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<Boresight>::select()
