//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalAtomsBase.h	--- CAL atom base class and atom collection class
//
//	DESCRIPTION:
//		CAL atoms are fundamental objects offering calibration-related
//		functionality (access to calibration data, performing calibration
//		algorithm on input data); CalAtom below is the base class from
//		which specialized classes are derived. Usually there is a 1:1
//		correspondence between a CalAtom and a CCF constituent; The former
//		is constructed from the latter
//
//		A CalAtomCollection is a container holding CalAtoms which differ
//		in CCD nr, node id, and mode id but have the same filterId and
//		accuracyLevel; CalAtomCollectionBase is the
//		underlying base class; it contains a fundamental method select()
//		to select a particular CAL atom from the container that matches
//		the input arguments (CCD, node, mode)
//
//	CHANGE HISTORY:
//		UL, 25/10/97: 1.0.0: - skeleton
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/CalAtomsBase.h,v 1.17 2003/12/19 17:17:36 ulammers Exp $
//
#ifndef CAL_ATOMS_BASE_H
#define CAL_ATOMS_BASE_H

#include "common.h"
#include "CalCommon.h"
#include "Xmm.h"
#include "Ccf.h"
#include "Dal.h"

#include <map>
#include <vector>

class CalAtom;	//	forward declaration

//
//	a collection of CalAtoms
//
class CalAtomCollectionBase {
public:
	virtual CalAtom *select(Ccd ccdNr, Node nodeNr,
							unsigned modeOffset = 0) const = 0;
		//
		//	PURPOSE:
		//		Select a concrete atom from the container
		//
		//	ARGUMENTS:
		//		ccdNr		: CCD nr (e.g. 1-7 for MOS)
		//		nodeNr		: node id
		//		modeOffset	: mode id as an unsigned started counting from 0;
		//					  0 is the default mode for the respective
		//					  instrument (e.g. FullFrame for MOS); see
		//					  Xmm.h for a list of modes and offsets
		//

	Filter filter() const { return _filter; }
	AccuracyLevel accuracy() const { return _accuracy; }
		//
		//	PURPOSE:
		//		Obtain filter id and accuracy level for this CalAtomCollection
		//


	size_t		entries() const { return _nEntries; }
		//
		//	PURPOSE:
		//		Give number of entries that this CalAtomCollection container
		//		holds
		//

	unsigned	ccds() const { return _nCCDs; }
	unsigned	nodes() const { return _nNodes; }
	unsigned	modes() const { return _nModes; }
		//
		//	PURPOSE:
		//		Returns numbers of different CCDs, nodes, modes that atoms
		//		in this collection refer to
		//

	virtual ~CalAtomCollectionBase() {}

protected:
	CalAtomCollectionBase() {}
	unsigned		index(unsigned modeNr) const;
	unsigned		index(unsigned ccdNr, unsigned nodeNr) const;
	unsigned		indexCcdMode(unsigned ccdNr, unsigned modeNr) const;
	unsigned		index(unsigned ccdNr, unsigned nodeNr, unsigned modeNr)
					const;
		//	PURPOSE:
		//		Get an index into the container pointing to the CalAtom
		//		that matches the input arguments
		//

	size_t			_nEntries;
	unsigned		_nCCDs, _nNodes, _nModes;
	Filter			_filter;		// filter setting
	AccuracyLevel	_accuracy;		// accuracy of atom collection
};


template<class AtomType>
class CalAtomCollection : public CalAtomCollectionBase {
public:
	//	empty default constructor
	CalAtomCollection() {}

	//	construct from CCF constituent
	CalAtomCollection(const XmmInstrument *instrument,
					  const Ccf *ccf,
					  Filter filter,
					  AccuracyLevel accuracy);

	//	select an atom from the collection
	AtomType *select(Ccd ccd, Node node,
					 unsigned modeOffset = 0) const;

	Ccd ccd() const { return _ccd; }
	Node node() const { return _node; }
	unsigned modeOffset() const { return _modeOffset; }

	~CalAtomCollection();

protected:
	std::vector<AtomType *>
						*_collection;	//	the collection
	mutable Ccd			_ccd;			//	selected CCD
	mutable Node		_node;			//	selected node
	mutable unsigned	_modeOffset;	//	selected modeOffset
};


//
//	collection destructor - always the same
//
template<class AtomType>
CalAtomCollection<AtomType>::~CalAtomCollection()
{
	for (unsigned n=0; n<_collection->size(); ++n)
		delete (*_collection)[n];	// delete the atom
	delete _collection;
}


class CalAtom {
public:
	virtual ~CalAtom() {}
	virtual void dump(ostream &) {
		errstr<<"Sorry - dumping to an ostream is not implemented for this "
				"CAL atom"<<Sas::error("MethodUnimplemented");
	}

protected:
	CalAtom() { _algoid = 0; }

	unsigned short	_algoid;
};

template<class AtomClass>
inline void calAtomUpcast(AtomClass *&ptr, CalAtom *atomPtr)
{
	ptr = reinterpret_cast<AtomClass *>(atomPtr);
} // calAtomUpcast

#endif
