//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CtiCorrector --- A class CtiCorrector is defined
//
//	DESCRIPTION:
//		Instances of class CtiCorrector can correct ADU channels for
//		the effect of Charge Transfer Inefficiency (CTI)
//
//	CHANGE HISTORY:
//		UL, 1999-06-15: 1.0.0: - first version from CalAtomsCommon
//		UL, 2000-03-27: 1.1.0: - added EMOS specific CTI corrector
//		UL, 2001-04-09: 1.2.0: - made <frame> argument of EPN CTI correction
//								 of type int32
//		UL, 2001-11-21: 1.9.0: - RGS HTR mode CTI correction method added
//
//	AUTHOR:
//		U. Lammers (UL), Thu Jun 17 11:42:35 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd et. al.
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/CtiCorrector.h,v 1.25 2006/06/06 12:34:43 rsaxton Exp $
//

#ifndef CTI_CORRECTOR_H
#define CTI_CORRECTOR_H

#include "STime.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalDataVector.h"
#include "Image.h"
#include "slatec++.h"

#include <vector>

//
//	---( CtiCorrection )--------------------------------------------------
//
class CtiCorrector : public CalAtom,
				  public virtual CalChild<CalAtomCollection<CtiCorrector> > {
public:
	CtiCorrector(const std::vector<vector<double> > &parallelCoefficients,
				 const std::vector<vector<double> > &serialCoefficients,
				 CalAtomCollection<CtiCorrector> *parent = 0)
		: CalChild<CalAtomCollection<CtiCorrector> >(parent),
		  _parallelCoefficients(parallelCoefficients),
		  _serialCoefficients(serialCoefficients) {}
		//
		//	PURPOSE:
		//		Construct CtiCorrector object
		//
		//	ARGUMENTS:
		//		parallelCoefficient: vector with parallel CTI coefficients;
		//							 one vector per node
		//		serialCoefficient  : vector with serial CTI coefficients;
		//							 one vector per node
		//


	static unsigned id() { return Ccf::CTI; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//

	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//

	virtual CalReal32Vector &ctiCorrect(const CalReal64Vector &time,
										const CalInt16Vector &rawX,
										const CalInt16Vector &rawY,
										const CalReal32Vector &eIn,
										CalReal32Vector &eOut,
										CalInt8Vector *nodes = 0) const;
		//
		//	PURPOSE:
		//		Convert gain corrected channels for the effect of CTI
		//                               RAWY               RAWX
		//			eOut = (1-parallelCti)    x (1-serialCti)    x Ein
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			rawX,
		//			rawY	: vector with rawX/rawY coordinates
		//			eIn		: vector with gain corrected channels to correct
		//			nodes	: node vector with nodes identifiers
		//					  in the range [0, MAX_NODES] signifying the
		//					  readout node which gave the pha channel
		//					  value passed via `eIn'
		//		Out:
		//			eOut	: vector with CTI corrected channels
		//
		//	RETURNS:
		//		vector with CTI corrected channels
		//

	const std::vector<vector<double> > &parallelCoefficients() const
			{ return _parallelCoefficients; }
	const std::vector<vector<double> > &serialCoefficients() const
			{ return _serialCoefficients; }
		//
		//	PURPOSE:
		//		Return references to principal data values
		//

	virtual ~CtiCorrector() {}
		//
		//	PURPOSE:
		//		Destroy CtiCorrector object
		//


	//
	//	--- I/F for specialized versions of above generic operations
	//

	//	EPN
	virtual CalReal32Vector &ctiCorrect(const CalReal64Vector &, // time
										const CalInt16Vector &,  // rawX
										const CalInt16Vector &,  // rawY
										const CalInt32Vector &,	 // frame
										const CalReal32Vector &, // eIn
										CalReal32Vector &eOut) const;

	//	RGS
	//	Spectroscopy
	virtual CalReal32Vector &ctiCorrect(const CalReal64Vector &,	// time
										const CalInt16Vector &,		// rawX
										const CalInt16Vector &,		// rawY
										const CalInt8Vector &,		// shape
										const CalInt16Vector &,		// separ.
										const CalReal32Vector &,	// eIn
										CalReal32Vector &eOut,			// eOut
										CalInt8Vector *) const		// nodes
		{ 
			errstr<<"Generic CTI corrector cannot perform RGS CTI correction"
				  <<Sas::error("UnsuitableCtiCorrector");
		  return eOut;
		}

	//	HTR
	virtual CalReal32Vector &ctiCorrect(const CalReal64Vector &,	// time
										const CalInt16Vector &,		// rawX
										int16,						// rawY
										const CalReal32Vector &,	// eIn
										CalReal32Vector &eOut,		// eOut
										CalInt8Vector *) const		// nodes
		{ 
			errstr<<"Generic CTI corrector cannot perform RGS HTR CTI "
					"correction"
				  <<Sas::error("UnsuitableCtiCorrector");
		  return eOut;
		}

protected:
	//	data members
	std::vector<vector<double> > _parallelCoefficients,
								 _serialCoefficients;
										//	vector with coefficients per nodes
};


//
//	a version of the CtiCorrector specialized for EMOS
// 
class EmosCtiCorrector : public virtual CtiCorrector {
public:
	EmosCtiCorrector(const CtiCorrector &genericCtiCorrector,
					 const Table *ctiTable, unsigned short algoid, 
                     const vector<SImage> colOffsets);
		//
		//	PURPOSE:
		//		Construct EmosCtiCorrector object
		//
		//	ARGUMENTS:
		//		genericCtiCorrector: generic CtiCorrector object (see above)
		//


	CalReal32Vector &ctiCorrect(const CalReal64Vector &time,
								const CalInt16Vector &rawX,
								const CalInt16Vector &rawY,
								const CalReal32Vector &eIn,
								CalReal32Vector &eOut,
								CalInt8Vector *nodes = 0) const;


	~EmosCtiCorrector();

private:
	STime	_refDate;
	double	_serialCtiRate,
			_parallelCtiRate;	
    vector<SImage> _colOffsets;

}; // EmosCtiCorrector



//
//	a version of the CtiCorrector specialized for EPN
// 
class EpnCtiCorrector : public virtual CtiCorrector {
public:
	EpnCtiCorrector(const CtiCorrector &genericCtiCorrector)
		: CalChild<CalAtomCollection<CtiCorrector> >
						(genericCtiCorrector.parent()),
		  CtiCorrector(genericCtiCorrector) {}
		//
		//	PURPOSE:
		//		Construct EpnCtiCorrector object
		//
		//	ARGUMENTS:
		//		genericCtiCorrector: generic CtiCorrector object (see above)
		//


	static void setCtiParams(const DataSet *set, const CifEntry *e);
		//
		//	PURPOSE:
		//		Set the static Cti parameters
		//


	CalReal32Vector &ctiCorrect(const CalReal64Vector &time,
								const CalInt16Vector &rawX,
								const CalInt16Vector &rawY,
								const CalReal32Vector &eIn,
								CalReal32Vector &eOut,
								CalInt8Vector *nodes = 0) const;


	CalReal32Vector &ctiCorrect(const CalReal64Vector &time,
								const CalInt16Vector &rawX,
								const CalInt16Vector &rawY,
								const CalInt32Vector &frame,
								const CalReal32Vector &eIn,
								CalReal32Vector &eOut) const;


	~EpnCtiCorrector();

private:
	static const unsigned N_CTI_PARAMETERS = 4;
	static struct CtiParam {
		unsigned		n;	//	number of parameter sets (should be 768)
		int				*iflParam;
		std::vector<float *>	ctiParam;
		//	0:	a
		//	1:	b
		//	2:	c
	} _ctiParam;

	static struct HotPixelParam {
		unsigned		n;	//	number of hot pixels
		std::vector<int *>	hotPixelParam;
		//	0:	quadrant	[0..3]
		//	1:	ccd			[0..2]
		//	2:	rawX		[0..63]
		//	3:	rawY		[0..199]
		//	4:	yLen		[0..199]
		float			*saturatedImp;	//	number of saturated impurities
	} _hotPixelParam;

	static struct DiscardedPixelParam {
		unsigned		n;	//	number of discarded pixels
		std::vector<int *>	discardedPixelParam;
	} _discardedPixelParam;

	static double			_ePerAdu,		//	electrons per ADU
							_keVPerE;		//	keV per electron

	static std::vector<unsigned short>
							_yShift;		//	y_shift parameter of CTI
	static std::vector<float>
							_ctiScale,		//	CTI scaling
							_eFFpar,		//	add. CTI parameters in eFF
							_modeFactor,	//	mode factor
							_lwPar,	 	    //	parameters in LW mode
							_swPar;		    //	parameters in SW mode

	static bool				_ctiParamsSet;
};


//
//	a version of the CtiCorrector specialized for RGS
// 
class RgsCtiCorrector : public virtual CtiCorrector {
public:
	RgsCtiCorrector(const CtiCorrector &genericCtiCorrector,
					const std::vector<double> &ctiX,
					const std::vector<double> &xCol,
					const std::vector<double> &ctiY,
					short unsigned algoid);
		//
		//	PURPOSE:
		//		Construct RgsCtiCorrector object
		//
		//	ARGUMENTS:
		//		genericCtiCorrector: generic CtiCorrector object (see above)
		//		xCol			   : grid on which ctiY is defined
		//		ctiY			   : vector with to be interpolated parallel
		//							 CTI coefficients; see [2] for details


	CalReal32Vector &ctiCorrect(const CalReal64Vector &time, // time
								const CalInt16Vector &rawX,  // rawX
								const CalInt16Vector &rawY,  // rawY
								const CalInt8Vector &shape,	 // shape
								const CalInt16Vector &separation,
								const CalReal32Vector &eIn, // eIn
								CalReal32Vector &eOut,
								CalInt8Vector *nodes = 0) const;
	CalReal32Vector &ctiCorrect(const CalReal64Vector &time,	// time
								const CalInt16Vector &rawX,		// rawX
								int16 rawY,						// rawY
								const CalReal32Vector &eIn,		// eIn
								CalReal32Vector &eOut,			// eOut
								CalInt8Vector *nodes = 0) const;// nodes
		//
		//	PURPOSE:
		//		CTI correct a vector of PHA values; second variant is
		//		for HTR mode
		//
		//	ARGUMENTS:
		//		time		: vector with event time tags (currently ignored)
		//		rawX/rawY	: event coordinates in PIXCOORD1 frame;
		//					  is scalar in HTR case to be set to
		//					  WINDOWDY/2+WINDOWY0
		//		shape		: shape code of ebent (currently ignored)
		//					  [irrelevant in HTR mode]
		//		separation	: number of transfers to the event preceding the
		//					  current one [irrelevant in HTR mode]
		//		eIn			: vector with uncorrected PHA values
		//		eOut		: vector with CTI corrected PHA values (also
		//					  returned)
		//		nodes		: optional vector with readout node identifiers
		//					  (if present, overrides usage of node state
		//					  variable)


	~RgsCtiCorrector();
		//
		//	PURPOSE:
		//		Destroy CTI corrector object constructed with above
		//		constructor(s).
		//

private:
	std::vector<double>						_xCti;
	Slatec::CubicHermiteInterpolant<double>	_yCti;

	//	ALGOID = 0
	CalReal32Vector &ctiCorrect0(const CalReal64Vector &time, // time
								 const CalInt16Vector &rawX,  // rawX
								 const CalInt16Vector &rawY,  // rawY
								 const CalInt8Vector &shape,	 // shape
								 const CalInt16Vector &separation,
								 const CalReal32Vector &eIn, // eIn
								 CalReal32Vector &eOut,
								 CalInt8Vector *nodes = 0) const;

	CalReal32Vector &ctiCorrect1(const CalReal64Vector &time, // time
								 const CalInt16Vector &rawX,  // rawX
								 const CalInt16Vector &rawY,  // rawY
								 const CalInt8Vector &shape,	 // shape
								 const CalInt16Vector &separation,
								 const CalReal32Vector &eIn, // eIn
								 CalReal32Vector &eOut,
								 CalInt8Vector *nodes = 0) const;
};

#endif
