//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	DarkFrameDataServer.cc --- implementation of DarkFrameDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the DarkFrameDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-07-28: 1.0.0: - first versin
//		UL, 2000-08-01: 1.1.0: - couple of bug fixes
//		UL, 2001-04-23: 1.2.0: - darkframe map is now of type real32
//
//	AUTHOR:
//		U. Lammers (UL), Fri Jul 28 13:54:22 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/DarkFrameDataServer.cc,v 1.13 2009/07/06 14:25:19 aibarra Exp $
//
#include "CalOalUtility.h"
#include "CalServer.h"
#include "DarkFrameDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( DarkFrameDataServer )------------------------------------------------
//
DarkFrameDataServer::DarkFrameDataServer(unsigned modeOffset,
										 const Proxy<SImage> &darkFrameImage,
									     const Table *table,
							 CalAtomCollection<DarkFrameDataServer> *parent)
	: CalChild<CalAtomCollection<DarkFrameDataServer> >(parent),
	  _modeOffset(modeOffset), _darkFrameImage(darkFrameImage),
	  _offX(0), _offY(0)
{
 /* The naming convention of the columns in the MOS darkframe CCFs has
    been corrupted (see CAL-SRN-0147-1-1 for part of the story) such
    that the ROW_OFFSET array goes with the COL_X0 and ROW_LEN columns
    and the COLUMN_OFFSET array with the ROW_Y0 and COL_LEN values */

	ColumnData *cdMode = table->column("MODE_ID")->data(),
			   //*cdColX0 = table->column("COL_X0")->data(),
			   *cdColX0 = table->column("ROW_Y0")->data(),
			   *cdColLen = table->column("COL_LEN")->data(),
			   *cdColOffset = table->column("COLUMN_OFFSET")->data(),
			   //*cdRowY0 = table->column("ROW_Y0")->data(),
			   *cdRowY0 = table->column("COL_X0")->data(),
			   *cdRowLen = table->column("ROW_LEN")->data(),
			   *cdRowOffset = table->column("ROW_OFFSET")->data();

	int16 *mode = cdMode->int16Data(),
		  *colX0 = cdColX0->int16Data(),
		  *colLen = cdColLen->int16Data(),
		  *colOffset = cdColOffset->int16Data(),
		  *rowY0 = cdRowY0->int16Data(),
		  *rowLen = cdRowLen->int16Data(),
		  *rowOffset = cdRowOffset->int16Data();

	unsigned N = table->rows();
	for (unsigned n=0; n<N; ++n)
		if (modeOffset == UNT(mode[n])) {
			int16 *start = colOffset+cdColOffset->position(n, colX0[n]);
			_offX = new vector<int16>(start, start+colLen[n]);
			start = rowOffset+cdRowOffset->position(n, rowY0[n]);
			_offY = new vector<int16>(start, start+rowLen[n]);
			break;
		}

	delete cdMode;
	delete cdColX0;
	delete cdColLen;
	delete cdColOffset;
	delete cdRowY0;
	delete cdRowLen;
	delete cdRowOffset;
} // DarkFrameDataServer::DarkFrameDataServer()


const vector<int16> &DarkFrameDataServer::rowOffsetVector() const
{
	if (!_offY)
		errstr<<"No row offset vector for mode "
			  <<Xmm::toMode(calServer.instrument()->modes()[_modeOffset])
			  <<Sas::error("NoRowOffsetVector");

	return *_offY;
} // DarkFrameDataServer::rowOffsetVector()


const vector<int16> &DarkFrameDataServer::columnOffsetVector() const
{
	if (!_offX)
		errstr<<"No column offset vector for mode "
			  <<Xmm::toMode(calServer.instrument()->modes()[_modeOffset])
			  <<Sas::error("NoRowOffsetVector");

	return *_offX;
} // DarkFrameDataServer::columnOffsetVector()


const SImage &DarkFrameDataServer::darkFrameMap() const
{
	return *_darkFrameImage;
} // DarkFrameDataServer::darkFrameMap()


DarkFrameDataServer::~DarkFrameDataServer()
{
	delete _offX;
	delete _offY;
} // DarkFrameDataServer::~DarkFrameDataServer()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<DarkFrameDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nModes		= instrument->nModes();
	_nEntries	= _nCCDs * _nModes;
	DarkFrameDataServer *null = 0;
	_collection = new vector<DarkFrameDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::DarkFrame);
	DataSet *dataSet = e->open();

	for (unsigned ccd=1; ccd<=_nCCDs; ++ccd) {
		//
		//	read dark frame map
		//
		string darkFrameMapName = "DARKFRAME_CCD" + toString(ccd);
		CcfConstituentArray ccfarr(dataSet, darkFrameMapName, 1);
		ArrayData *ad = ccfarr.block()->data();
		vector<unsigned long> extents = ad->parent()->size();
		Proxy<SImage> darkFrameMap(new SImage(extents[0], extents[1],
											  ad->int16Data()));
		if (ccfarr.algoid() == 1) {
			//	ALGOID-1 darkframe are scaled by factor 100
			unsigned N = darkFrameMap->xExtent() * darkFrameMap->yExtent();
			real32 *a = darkFrameMap->data();
			for (unsigned n=0; n<N; a[n++]/=100.) ;
		}

		delete ad;

		for (unsigned mode=0; mode<_nModes; ++mode) {
			DarkFrameDataServer *&qeds =
									(*_collection)[indexCcdMode(ccd, mode)];
			string offsetTableName = "OFFSET_CCD" + toString(ccd);
			CcfConstituentTable ccftab(dataSet, offsetTableName);
			qeds = new DarkFrameDataServer(mode, darkFrameMap,
										   ccftab.block(),
										   this);
		}
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<DarkFrameDataServer>::CalAtomCollection()


/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
DarkFrameDataServer *CalAtomCollection<DarkFrameDataServer>::select(
					Ccd ccd,
					Node,
					unsigned modeOffset) const
{
	return (*_collection)[indexCcdMode(ccd, modeOffset)];
} // CalAtomCollection<DarkFrameDataServer>::select()
