//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2020 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	FWCDataServer.cc --- implementation of class to return data from
//			     the Filter Wheel Closed data CCF
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 22-09-16: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2016-09-22
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//

#include "FWCDataServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

typedef multimap<const pair<int,unsigned>,FWCevent>::iterator mapIter;

// Define the FWCevents class
// 
FWCevent::FWCevent(double time, short rawx, short rawy, short detx, short dety,
           short pi, int flag, unsigned pattern, unsigned ccd,
           float weight, int rev) : 
          _time(time), _rawx(rawx), _rawy(rawy), _detx(detx), _dety(dety),
          _pi(pi), _flag(flag), _pattern(pattern), _ccd(ccd),
          _weight(weight), _rev(rev)
{
}

FWCevent:: ~FWCevent()
{
  PROC;
     // Nothing so far
  ENDPROC;
};

/* Method: return time */
/*
double FWCevent::time()
{ return _time; }
short FWCevent::rawx()
{return _rawx;}
short FWCevent::rawy()
{return _rawy;}
short FWCevent::detx()
{return _detx;}
short FWCevent::dety()
{return _dety;}
short FWCevent::pi()
{return _pi;}
int FWCevent::flag()
{return _flag;}
unsigned FWCevent::pattern()
{return _pattern;}
unsigned FWCevent::ccd()
{return _ccd;}
float FWCevent::weight()
{return _weight;}
int FWCevent::rev()
{return _rev;}
*/



//
//	----------------------( mode parameters )-----------------------------
//

//  Define the FWCDataServer class
//
FWCDataServer::FWCDataServer( multimap<const pair<int,unsigned>, 
                                           FWCevent> fwcevents,
                              FWCexptime etime,
                              CalAtomCollection<FWCDataServer> *parent)
        : CalChild<CalAtomCollection<FWCDataServer> >(parent),
          _fwcevents(fwcevents),_et(etime)
{
}
// Return events from these CCDs and Revolutions
vector<FWCevent> FWCDataServer::getEvents(multimap<int, unsigned> revCcd)
{
  vector<FWCevent> fwcevout;

 // Loop over each Revolution_CCD pair
  for (multimap<int, unsigned>::iterator it = revCcd.begin();
       it != revCcd.end(); ++it)
  {
   // Extract the Rev,Ccd pair
     pair<int,unsigned> p = make_pair((*it).first,(*it).second);

   // Extract all events from this rev,ccd combination
     mapIter s_it;
     pair<mapIter, mapIter> keyRange = _fwcevents.equal_range(p);
     for (s_it = keyRange.first;  s_it != keyRange.second;  ++s_it)
        {
           fwcevout.push_back((*s_it).second);
        }
  }
  return fwcevout;
}

//
// Define the entry point
// 
template<>
CalAtomCollection<FWCDataServer>::CalAtomCollection(
	                                     const XmmInstrument *instrument,
					     const Ccf *ccf,
					     Xmm::Filter filter,
					     AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
        _nCCDs		= instrument->nCcds();

	FWCDataServer *null = 0;
	_collection = new vector<FWCDataServer *>(1, null);

	CifEntry *e = ccf->entry(instrument, Ccf::FWC);
	DataSet *dataSet = e->open();

      // Depending on the instrument open the correct block(s)
        Table *table = dataSet->table(e->queryBlockName("EVENTS"));

      // Fill vectors - may not be the most efficient - revise !
        vector<double> time;
        vector<float> wt;
        vector<short> rawx, rawy, detx, dety,pi;
        vector<int> flag,rev;
        vector<unsigned> pattern,ccd;

        CalOalUtility::fillVector(table, "TIME", time);
        CalOalUtility::fillVector(table, "RAWX", rawx);
        CalOalUtility::fillVector(table, "RAWY", rawy);
        CalOalUtility::fillVector(table, "DETX", detx);
        CalOalUtility::fillVector(table, "DETY", dety);
        CalOalUtility::fillVector(table, "PI", pi);
        CalOalUtility::fillVector(table, "FLAG", flag);
        CalOalUtility::fillVector(table, "PATTERN", pattern);
        CalOalUtility::fillVector(table, "CCDNR", ccd);
        CalOalUtility::fillVector(table, "WEIGHT", wt);
        CalOalUtility::fillVector(table, "REV", rev);

      // Produce a map with all of the events
       multimap<const pair<int,unsigned>, FWCevent> m;
       pair<int,unsigned> p = make_pair(0,0);
   
      // Loop over each event
       for (unsigned i=0; i<table->rows(); ++i) {      

         // Make the key
           p = make_pair(rev[i],ccd[i]);

         // Insert the event into the multimap.
         // mymap.insert(make_pair("Group1",  1));
           m.insert(make_pair(p, FWCevent(
                        time[i],rawx[i],rawy[i],detx[i],dety[i],pi[i],
                        flag[i],pattern[i],ccd[i],wt[i],rev[i])
                   ));
       }

      /* Read in the exposure time per ccd and per revolution */
        Table *etable = dataSet->table(e->queryBlockName("EXPTIME"));
        
        FWCexptime etime;

        CalOalUtility::fillVector(etable, "REV", etime.rev);
        CalOalUtility::fillVector(etable, "TSTART", etime.tstart);
        CalOalUtility::fillVector(etable, "TSTOP", etime.tstop);
        CalOalUtility::fillVector(etable, "CCDNR", etime.ccd);
        CalOalUtility::fillVector(etable, "EXPOSURE", etime.exptime);
        CalOalUtility::fillVector(etable, "NFLAG", etime.nflag);

     // Create the collection - just one
	(*_collection)[0] = new FWCDataServer(m, etime, this);

      // Delete the column pointers
      /*
        delete cdmodeId;
        delete cdngrp;
        delete cdflag;
	delete cdgmodId;
	delete cdsmin;
	delete cdsmax;
	delete cdomin;
	delete cdomax;
        //delete cdenergy;
        //delete cdres;
      */
	dataSetServer->close(dataSet);
} // CalAtomCollection<FWCDataServer>::CalAtomCollection


/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
FWCDataServer *CalAtomCollection<FWCDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
        if (!(*_collection)[0])
           errstr<<"The CCF does not contain Filter Wheel Closed data "
                          <<Sas::error("NoData");

	return (*_collection)[0];
} // CalAtomCollection<FWCDataServer>::select
