//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	GratingDataServer.cc --- implementation of GratingDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the GratingDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//		UL, 2000-04-13: 1.1.0: - bug fix in ~GratingDataServer()
//		UL, 2000-05-05: 1.2.0: - bug fix in reading of reflection efficiencies
//		UL, 2000-08-31: 1.3.0: - bug fix in alpha(): F/L factor not considered
//		UL, 2000-11-14: 1.4.0: - modifications following latest changes in [2]:
//		UL, 2000-11-17: 1.5.0: - added
//									GratingDataServer::scatteringDistribution()
//		UL, 2000-12-04: 1.6.0: - bug fix: wrong sign in alpha() to compute
//								 incidence angle from TELCOORD(theta, phi):
//								 now the Ae is largest at phi=0 (~CCD 5 for
//								 EMOS1)
//		UL, 2000-12-08: 1.7.0: - change in misalignmentFigureDistribution():
//								 returned spectrum is now centered on first
//								 moment mu0 and mu0-floor(mu0)==0
//		UL, 2001-01-10: 1.8.0: - added GratingDataServer::chi()
//		UL, 2001-01-12: 1.9.0: - alpha0 did not include rotation of RGA
//								 about +Y
//		UL, 2001-02-16: 1.10.0:- new method: order() to compute spectral
//								 order from alpha/beta/E
//							   - scatteringDistribution() changes:
//									* revised I/F
//									* number of channels in returned spectrum
//									  is now odd with betaM in central channel
//									* bug fixes (was also wrong in [2])
//		UL, 2001-02-19: 1.11.0:- scatteringDistribution(): I/F roll-back
//								 (betaM, alpha, E) -> (alpha, order, E)
//		UL, 2001-02-20: 1.12.0:- added bowingFigureDistribution()
//		UL, 2001-03-09: 1.12.1:- bug fix: declaration of _Flarge/_Fsmall in
//								 GratingDataServer::scatteringDistribution()
//								 shadowed class variables
//		UL, 2001-03-12: 1.11.0:- added version of
//									scatteringDistribution()
//								 which accepts number of output spectrum
//								 channels explicitely
//		UL, 2001-03-28: 1.12.0:- broadeningFudge->broadeningFactor
//		UL, 2001-05-04: 1.13.0:- added
//									broadeningDistribution()
//		UL, 2001-10-24: 1.14.0:- support for multiplicative empirical
//								 correction curves in
//									o selfVignetting
//									o reflectionEfficiencies
//								 this is driven by tables in RGS?_QUANTUMEF;
//								 backwards compatible; missing data trigger
//								 instantiation of constant 1-curve
//		UL, 2002-01-25: 1.15.0:- changed way in which RGA_EFFCORR data
//								 are applied to reflection efficiencies: now
//								 data are first interpolated and then applied
//								 to the efficiency curves at the original
//								 energy (lambda) grid
//		UL, 2002-01-29: 1.15.1:- fixed problem introduced with 1.15.0:
//								 neglected case of missing reflection
//								 efficiency correction data (old-style
//								 QUANTUMEF CCFs)
//		UL, 2002-06-20: 1.16.0:- added ICO arguments to
//									misalignmentFigureDistribution()
//									bowingFigureDistribution()
//								 needed for extended source support in
//								 rgsrmfgen
//		UL, 2002-06-26: 1.16.1:- bug fix in misalignmentFigureDistribution()
//								 related to recent addition of ICO argument
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 18:24:59 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd et. al
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/GratingDataServer.cc,v 1.40 2012/12/18 15:45:05 aibarra Exp $
//
#include <cmath>
#include <cstdlib>
#include <algorithm>

#include "Tracer.h"
#include "CalOalUtility.h"
#include "Celestial.h"
#include "CalCommon.h"
#include "slatec++.h"
#include "PhysicalConstants.h"
#include "CalServer.h"
#include "MiscDataServer.h"
#include "GratingDataServer.h"

using namespace CalOalUtility;
using namespace std;

//
//	---( GratingDataServer )--------------------------------------------------
//
GratingDataServer::GratingDataServer(
							const CcfConstituentTable &figureTable0,
							const CcfConstituentTable &figureTable1,
							const CcfConstituentTable &scatterTable,
						    const CcfConstituentTable &laScatterTable,
							const CcfConstituentTable &reflectionTable,
							const CcfConstituentTable &reflectionCorrTable,
							const CcfConstituentTable &selfVignCorrTable,
							const CcfConstituentTable &interceptTable,
							const CcfConstituentTable &obscurationTable,
							double alpha0,
							double grooveSpacing,
							double FoverL,
							CalAtomCollection<GratingDataServer> *parent)
	: CalChild<CalAtomCollection<GratingDataServer> >(parent),
	  _alpha0(alpha0), _grooveSpacing(grooveSpacing), _FoverL(FoverL),
	  _Fsmall(0.), _Flarge(0.)
{
	//
	//	parameters for misalignmentFigureDistribution
	//
	if (figureTable0.algoid() > 0) {
		CalOalUtility::fillVector(figureTable0.block(), "AMPL", _ampl[0]);
		CalOalUtility::fillVector(figureTable0.block(), "CENTER", _center[0]);
		CalOalUtility::fillVector(figureTable0.block(), "SIGMA", _sigma[0]);
	}

	//
	//	scale center + width of Gaussians with broadening factor
	//	(can be called "FUDGE" exclusive-or "SCALE" in CCF)
	//
	unsigned nGaus = _ampl[0].size();
	double broadeningFactor =
			CalOalUtility::optionalAttribute(figureTable0.block(), "FUDGE",1.)*
			CalOalUtility::optionalAttribute(figureTable0.block(), "SCALE",1.);

	for (unsigned n=0; n<nGaus; ++n) {
		_center[0][n] *= broadeningFactor;
		_sigma[0][n] *= broadeningFactor;
	}

	//
	//	parameters for bowingFigureDistribution
	//
	if (figureTable1.block()) {
		//	optional - older LineSpreadFunc CCF have now BOWS table
		CalOalUtility::fillVector(figureTable1.block(), "AMPL", _ampl[1]);
		CalOalUtility::fillVector(figureTable1.block(), "CENTER", _center[1]);
		CalOalUtility::fillVector(figureTable1.block(), "SIGMA", _sigma[1]);
	} else
		errstr<<"CCF::LineSpreadFunc contains no BOWS table - "
				"bowingFigureDistribution() will return empty spectrum"
			  <<Sas::warning("MissingTableInCcf");

	//
	//	scattering parameters
	//
	CalOalUtility::fillVector(scatterTable.block(), "A", 0, _scatterA);
	CalOalUtility::fillVector(scatterTable.block(), "B", 0, _scatterB);
	CalOalUtility::fillVector(scatterTable.block(), "RMS", 0, _scatterRMS);
	CalOalUtility::fillVector(scatterTable.block(), "CORLEN", 0, _scatterLcor);

	vector<double>	lambdaGrid, rms;
	CalOalUtility::fillVector(laScatterTable.block(), "WAVELENG", lambdaGrid);
	CalOalUtility::fillVector(laScatterTable.block(), "RMS", rms);
	_largeAngleScatterRMS = InterpolatedCurve(lambdaGrid, rms);

	//
	//	read reflection efficiency data
	//
	ColumnData *cdAlpha = reflectionTable.block()->column("ALPHA")->data(),
			   *cdLambda = reflectionTable.block()->column("LAMBDA")->data();
	unsigned nOrders = maxOrder - minOrder + 1;
	vector<const ColumnData *> cdEff(nOrders);
	for (unsigned n=0; n<nOrders; ++n)
		cdEff[n] =
			reflectionTable.block()->column("EFF" +
											toString(n+maxOrder))->data();

	real32 *alpha = cdAlpha->real32Data(),
		   *lambda = cdLambda->real32Data();
	vector<const real32 *> eff(nOrders);
	for (unsigned n=0; n<nOrders; ++n)
		eff[n] = cdEff[n]->real32Data();

	unsigned N = reflectionTable.block()->rows(),
			 Ne;
	for (Ne=0; Ne<N; ++Ne) {
		if (Ne && lambda[Ne] == lambda[0])
			break;
		_reflectionEgrid.push_back(PhysicalConstants::eVtimesAo/lambda[Ne]);
	}
	sort(_reflectionEgrid.begin(), _reflectionEgrid.end());

	//	check if number of table rows is multiple of egrid vector size
	if (N % Ne)
		errstr<<"Number of rows in table `"
			  <<reflectionTable.block()->qualifiedName()
			  <<"' is not a multiple of "<<Ne<<" as expected"
			  <<Sas::error("DataIntegrityError");

	unsigned Nalpha = N/Ne;
	_alphaGrid.resize(Nalpha);
	for (unsigned n=0; n<Nalpha; ++n)
		_alphaGrid[n] = alpha[n*Ne] * DEG_TO_RAD;

	_reflectionEfficiencies = Image<InterpolatedCurve *>(nOrders, Ne);
	for (unsigned i=0; i<nOrders; ++i)
		for (unsigned j=0; j<Ne; ++j) {
			vector<double> effVec(Nalpha);
			for (unsigned k=0; k<Nalpha; ++k)
				//	wavelength increases (E decreases) with increasing row
				//	number
				effVec[k] = eff[i][(k+1)*Ne-j-1];
			_reflectionEfficiencies[i][j] =
					new InterpolatedCurve(_alphaGrid, effVec);
		}

	delete cdAlpha;
	delete cdLambda;
	for (unsigned n=0; n<nOrders; delete cdEff[n++]) ;

	//
	//	reflection efficiency corrections
	//
	const Table *cb = reflectionCorrTable.block();
	if (cb) {
		//	table is present - read contents
		assert(cb->column("LAMBDA")->units() == "angstrom");
		vector<double> &eGrid = _reflectionEfficienciesCorrEgrid;
		CalOalUtility::fillVector(cb, "LAMBDA", eGrid);
		if (eGrid[0] == 0.)
			errstr<<"Cannot have lambda=0 - the energy in the universe is "
					"finite"<<Sas::error("ZeroWavelength");
		if (!CalCommon::sorted(eGrid))
			errstr<<"`"<<cb->column("LAMBDA")<<"' is not sorted in ascending "
					"order"<<Sas::error("DataUnsorted");
		//	revsere eGrid and transform to eV
		std::reverse(eGrid.begin(), eGrid.end());
		std::transform(eGrid.begin(), eGrid.end(), eGrid.begin(),
					   CalFunctors::Angstrom2eV<double>());
		for (unsigned n=0; n<nOrders; ++n) {
			vector<double> v;
			CalOalUtility::fillVector(cb, "CORR" + toString(n), v);
			std::reverse(v.begin(), v.end());
			_reflectionEfficienciesCorrValues.push_back(
				new InterpolatedCurve(_reflectionEfficienciesCorrEgrid, v));
		}
	}

	//
	//	self vignetting + correction curve
	//
	_betaVign =
		reflectionTable.block()->attribute("BETA_VIG")->asReal() * DEG_TO_RAD;
	cb = selfVignCorrTable.block();
	if (cb) {
		//	table is present - read contents
		assert(cb->column("BETA")->units() == "rad");
		std::vector<double> beta, f;
		CalOalUtility::fillVector(cb, "BETA", beta);
		CalOalUtility::fillVector(cb, "FACTOR", f);
		_selfVignettingCorr = InterpolatedCurve(beta, f);
	} else
		//	table can be missing from early CCFs - register constant
		//	correction curve
		_selfVignettingCorr = InterpolatedCurve(1.);

	//
	//	read intercept efficiency data
	//
	ColumnData *cdE = interceptTable.block()->column("ENERGY")->data(),
			   *cdf = interceptTable.block()->column("FRACTION")->data();

	N = interceptTable.block()->rows();
	vector<double> eGridIntercept(N),
				   factIntercept(N);
	real32 *E = cdE->real32Data(),
		   *f = cdf->real32Data();
	for (unsigned n=0; n<N; ++n) {
		eGridIntercept[n] = E[n];
		factIntercept[n] = f[n];
	}
	delete cdE;
	delete cdf;
	_interceptEfficiency = InterpolatedCurve(eGridIntercept,
											 factIntercept);

	//
	//	read obscuration data
	//
	cdE = obscurationTable.block()->column("ENERGY")->data();
	cdf = obscurationTable.block()->column("FRACTION")->data();

	N = obscurationTable.block()->rows();
	vector<double> eGridObscure(N),
				   factObscure(N);
	E = cdE->real32Data();
	f = cdf->real32Data();
	for (unsigned n=0; n<N; ++n) {
		eGridObscure[n] = E[n];
		factObscure[n] = f[n];
	}
	delete cdE;
	delete cdf;
	_obscurationEfficiency = InterpolatedCurve(eGridObscure,
											   factObscure);
	} // GratingDataServer::GratingDataServer()


void GratingDataServer::dump(ostream &)
{
	errstr<<"Sorry dumping GratingDataServer not yet implemented"
		  <<Sas::warning("NotYetImplemented");
} // GratingDataServer::dump()


InterpolatedCurve GratingDataServer::reflectionEfficiency(double theta,
														  double phi,
														  int order) const
{
	return reflectionEfficiency(alpha(theta, phi), order);
} // GratingDataServer::reflectionEfficiency()


InterpolatedCurve GratingDataServer::reflectionEfficiency(double alphaIn,
														  int order) const
{
	if (order<minOrder || order>maxOrder)
		errstr<<"Spectral order ("<<order<<") out of bounds - must be "
				"in interval ["<<minOrder<<", "<<maxOrder<<']'
			  <<Sas::error("OutOfBounds");

	order = abs(order);
	unsigned N = _reflectionEgrid.size();
	vector<double> v(N);
	for (unsigned n=0; n<N; ++n)
		v[n] = (*_reflectionEfficiencies[order][n])(alphaIn);

	InterpolatedCurve eff(_reflectionEgrid, v);

	if (_reflectionEfficienciesCorrValues.size() > UNT(order))
		//	might be missing (old-style RGS_QUANTUMEF CCF)
		eff *= *_reflectionEfficienciesCorrValues[order];
#if 0
	//	old algorithm
	N = _reflectionEfficienciesCorrEgrid.size();
	if (N) {
		vector<double> vadd(N);
		for (unsigned n=0; n<N; ++n)
			vadd[n] = eff(_reflectionEfficienciesCorrEgrid[n]) *
							_reflectionEfficienciesCorrValues[order][n];
		eff.supplement(_reflectionEfficienciesCorrEgrid, vadd);
	}
#endif

	return eff;
} // GratingDataServer::reflectionEfficiency()


InterpolatedCurve GratingDataServer::interceptEfficiency(double theta,
														 double phi) const
{
	InterpolatedCurve ie(_interceptEfficiency);
	return ie *	fabs(alpha(theta, phi)/_alpha0);
} // GratingDataServer::interceptEfficiency()


InterpolatedCurve GratingDataServer::obscurationEfficiency(double theta,
														   double phi) const
{
	InterpolatedCurve oe(_obscurationEfficiency);
	return oe *	fabs(alpha(theta, phi)/_alpha0);
} // GratingDataServer::obscurationEfficiency()


double GratingDataServer::obscurationEfficiency(double E,
											    double theta,
											    double phi) const
{
	return _obscurationEfficiency(E) * fabs(alpha(theta, phi)/_alpha0);
} // GratingDataServer::obscurationEfficiency()


vector<double> GratingDataServer::selfVignettingEfficiency(
								const vector<double> &dispersionAngles) const
{
	unsigned N = dispersionAngles.size();
	vector<double> v(N);
	for (unsigned n=0; n<N; ++n)
		v[n] = selfVignetting(dispersionAngles[n]);

	return v;
} // GratingDataServer::selfVignettingEfficiency()


vector<double> GratingDataServer::selfVignettingEfficiency(
													const vector<double> &E,
													double theta,
													double phi,
													int order) const
{
	unsigned N = E.size();
	vector<double> v(N);
	double alphaIn = alpha(theta, phi);
	for (unsigned n=0; n<N; ++n)
		v[n] = selfVignetting(beta(alphaIn, order, E[n]));

	return v;
} // selfVignettingEfficiency()


//
//	return incidence angle for TELCOORD position (theta/phi) in ["/rad]
//
double GratingDataServer::alpha(double theta, double phi) const
{
	return _alpha0 - _FoverL * theta/3600.*DEG_TO_RAD * cos(phi);
} // GratingDataServer::alpha()


double GratingDataServer::chi(double theta, double phi) const
{
	return  -asin(theta/3600.*DEG_TO_RAD*sin(phi));
} // GratingDataServer::chi()


double GratingDataServer::selfVignetting(double beta) const
{
	return (beta < _betaVign ? 1. : _betaVign/tan(beta)) *
			_selfVignettingCorr(beta);
} // GratingDataServer::selfVignetting()


Spectrum GratingDataServer::misalignmentFigureDistribution(
										double nBetaHWHM,
										double betaBinWidth,
										double chromaticMagnification,
										double inChannelOffset) const
{
	return gaussianFigureDistribution(nBetaHWHM, betaBinWidth,
									  chromaticMagnification,
									  inChannelOffset,
									  _ampl[0], _center[0], _sigma[0]);
} // GratingDataServer::misalignmentFigureDistribution()


Spectrum GratingDataServer::bowingFigureDistribution(
										double nBetaHWHM,
										double betaBinWidth,
										double chromaticMagnification,
										double inChannelOffset) const
{
	return gaussianFigureDistribution(nBetaHWHM, betaBinWidth,
									  chromaticMagnification,
									  inChannelOffset,
									  _ampl[1], _center[1], _sigma[1]);
} // GratingDataServer::bowingFigureDistribution()


Spectrum GratingDataServer::gaussianFigureDistribution(
										double nBetaHWHM,
										double betaBinWidth,
										double chromaticMagnification,
										double inChannelOffset,
										const vector<double> &amplVal,
										const vector<double> &centerVal,
										const vector<double> &sigmaVal) const
{
	double betaSigma = 2.*nBetaHWHM/M_2SQ2LN2;
	vector<double> ampl(amplVal),
				   center(centerVal),
				   sigma(sigmaVal);
	unsigned nGauss = ampl.size();
	for (unsigned n=0; n<nGauss; ++n) {
		ampl[n] *= M_SQRT_TWO_PI * sigma[n];
	}

	//	total normalization
	double norm = 0.;
	for (unsigned n=0; n<nGauss; ++n)
		norm += ampl[n];

	//	calculate first moment and renormalize individual amplitudes
	double firstMoment = 0.;	//	in beta space
	for (unsigned n=0; n<nGauss; ++n) {
		ampl[n] /= norm;
		firstMoment += ampl[n]*center[n];
	}

	//	apply chromatic magnification correction to centers and sigmas
	double CM = 1. + chromaticMagnification,
		   maxSigma = 0;
	for (unsigned n=0; n<nGauss; ++n) {
		center[n] = CM*(center[n]-firstMoment) + firstMoment;
												//	distribution is centered
												//	on first moment mu0
		sigma[n] *= CM;
		if (sigma[n] > maxSigma)
			maxSigma = sigma[n];
	}

	//	go to channel space - also adjust the centers such that the
	//	first moment lies at <inChannelOffset> from the array center
	unsigned nBetaBins = 1+ 2 * UNT(maxSigma/betaBinWidth * betaSigma + .5);
	double betaBins = DBL(nBetaBins),
		   betaSpan = betaBins * betaBinWidth,
		   chanPerRad = betaBins/betaSpan,
		   firstMomChan = firstMoment * chanPerRad,
		   deltaCenter = inChannelOffset - firstMomChan;
	for (unsigned n=0; n<nGauss; ++n) {
		center[n] = center[n]*chanPerRad + deltaCenter;
		sigma[n] *= chanPerRad;
	}

	//
	//	fill the spectrum
	//
	Spectrum spec(nBetaBins, 0.);
	int iBetaBins = nBetaBins/2;
	for (int ii=1; ii>=-1; ii-=2)
		for (int i=0; i<=iBetaBins; ++i) {
			int indx = ii*i;
			spec[indx+iBetaBins] = 0.;
			for (unsigned n=0; n<nGauss; ++n) {
				double oneOversq2sigma = 1./(M_SQRT_2*sigma[n]),
					   argPlus = DBL(indx)+.5-center[n],
					   argMinus = DBL(indx)-.5-center[n];
				spec[indx+iBetaBins] += .5*ampl[n] *
							fabs(Slatec::erf(oneOversq2sigma * argPlus) -
								 Slatec::erf(oneOversq2sigma * argMinus));
			}
		}

	//	set first moment
	spec.firstMoment(iBetaBins + inChannelOffset);
	SAS_DEBUG({
		unsigned NN = spec.size();
		double ntot = 0.;
		for (unsigned n=0; n<NN; ntot+=spec[n++]);
		cout<<NN<<' '<<ntot<<' '<<spec.firstMoment()<<endl;
	})

	return spec;
} // GratingDataServer::gaussianFigureDistribution()


Spectrum GratingDataServer::scatteringDistribution(unsigned nChannelsLeft,
												   unsigned nChannelsRight,
												   double betaChannelWidth,
												   double alpha,
												   int order,
												   double E,
												   bool withSmallAngleContrib,
												   bool withLargeAngleContrib)
const
{
	double lambda = PhysicalConstants::eVtimesAo/E,		//	[Ang]
		   betaOut = beta(alpha, order, E),				//	[rad]
		   k = M_TWO_PI/lambda,							//	[1/Ang]
		   sigmaSmall = _scatterRMS[0],					//	[Ang]
		   sigmaSmall2 = sigmaSmall * sigmaSmall,
		   sigmaLarge = _largeAngleScatterRMS(lambda),	//	[Ang]
		   sigmaLarge2 = sigmaLarge * sigmaLarge,
		   lSmall = 1.e4 * _scatterLcor[0],				//	um -> Ang
		   lSmall2 = lSmall * lSmall,
		   lLarge = 1.e4 * _scatterLcor[1],				//	um -> Ang
		   lLarge2 = lLarge*lLarge;

	double sinBeta = sin(betaOut),
		   cosBeta = cos(betaOut),
		   sinAlpha = sin(alpha),
		   sinAlphaSinBeta = sinAlpha + sinBeta,
		   x = 1./(k*lLarge*sinBeta*sinBeta),
		   x2 = x*x,
		   x3 = x2*x,
		   x4 = x2*x2;

   _Fsmall = 1.-exp(-SQR(k*sigmaSmall*sinAlphaSinBeta));
   _Flarge = 1.-exp(-SQR(k*sigmaLarge*sinAlphaSinBeta));

	double R = (_scatterA[0] + _scatterA[1]*x + _scatterA[2]*x2 +
			    _scatterA[3]*x3 + _scatterA[4]*x4)/
			   (_scatterB[0] + _scatterB[1]*x + _scatterB[2]*x2 +
			    _scatterB[3]*x3 + _scatterB[4]*x4);

	double WsmallFac = M_SQRT_2/M_SQRT_TWO_PI*lSmall*sigmaSmall2,
		   WlargeFac = lLarge/M_PI*sigmaLarge2,
		   C = betaChannelWidth*k*sinBeta/(
								sinAlphaSinBeta*sinAlphaSinBeta*
							    sinAlphaSinBeta*sinAlphaSinBeta),
		   Clarge =	withLargeAngleContrib
						? R*_Flarge*WlargeFac/sigmaLarge2 : 0.,
		   Csmall =	withSmallAngleContrib
						? (1.-_Flarge)*_Fsmall*WsmallFac/sigmaSmall2 : 0.;

	//
	//	fill the spectrum
	//
	int nChannels[] = { nChannelsRight, nChannelsLeft };
	Spectrum spec(nChannelsRight+nChannelsLeft+1, 0.);
	for (int ii=1; ii>=-1; ii-=2) {
		int iii = (1-ii)/2;
		for (int i=0; i<=nChannels[iii]; ++i) {
			int indx = ii*i;
			double beta = betaOut + DBL(indx) * betaChannelWidth,
				   sinBeta = sin(beta),
				   sinAlphaBeta = sinAlpha+sinBeta,
				   sinAlphaBeta2 = sinAlphaBeta*sinAlphaBeta,
				   p = k*(cosBeta - cos(beta)),
				   p2 = p*p;

			spec[indx+nChannelsLeft] = C*sinAlphaBeta2*sinAlphaBeta2 * (
									   Clarge / (1.+lLarge2*p2) +
									   Csmall * exp(-lSmall2*p2));
//
//	do not consider un-scattered light in central channel
//			if (indx == 0)
//				spec[iBetaBins] += (1.-_Flarge)*(1.-_Fsmall);
		}
	}

	SAS_DEBUG(
	cout<<spec.norm()<<" + "
		<<(1.-_Flarge)*(1.-_Fsmall)<<" = "
		<<spec.norm() + (1.-_Flarge)*(1.-Fsmall)<<endl;
	)

	return spec;
} // GratingDataServer::scatteringDistribution()


Spectrum GratingDataServer::scatteringDistribution(double nBetaHWHM,
												   double betaChannelWidth,
												   double alpha,
												   int order,
												   double E,
												   bool withSmallAngleContrib,
												   bool withLargeAngleContrib)
const
{
	double betaSigma = 2.*nBetaHWHM/M_2SQ2LN2;
	const double SIGMA	= 5.e-4;	//	typical width in [rad]
	unsigned nChanLeft = UNT(SIGMA/betaChannelWidth * betaSigma + .5),
			 nChanRight = nChanLeft;

	return scatteringDistribution(nChanLeft, nChanRight,
								  betaChannelWidth,
								  alpha, order, E,
								  withSmallAngleContrib,
								  withLargeAngleContrib);
} // GratingDataServer::scatteringDistribution()


//
//	following implementation taken from calib.cc in rgsrmfgen-0.40
//
Spectrum GratingDataServer::broadeningDistribution(double stride,
												   double dLambda,
												   double betaM,
												   int order) const
{
	Spectrum dist;

	if (order==0 || dLambda==0.)
		return dist;

	double power = stride * _grooveSpacing * sin(betaM)/fabs(order * dLambda);
	if (power >= 1)
		return dist;

	int n = 1 + 2 * (int)floor(.5 * (1./power - 1.));

	dist.resize(n + 2);
	dist[0] = dist[n+1] = .5 * (1. - n * power);
	for (Spectrum::iterator itr = dist.begin()+1; n--; *itr++ = power) ;

	return dist;
} // GratingDataServer::broadeningDistribution()


double GratingDataServer::beta(double alpha, int order, double E) const
{
	return acos(order*PhysicalConstants::eVtimesAo/E/_grooveSpacing +
				cos(alpha));
} // GratingDataServer::beta()

double GratingDataServer::lambda2beta(double alpha, int order, double lambda) const
{
	return acos(DBL(order)*lambda/_grooveSpacing + cos(alpha));
} // GratingDataServer::beta()

double GratingDataServer::order(double alpha, double beta, double E) const
{

	return _grooveSpacing/(PhysicalConstants::eVtimesAo/E)*
			(cos(beta)-cos(alpha));
} // GratingDataServer::order()


double GratingDataServer::energy(double alpha, double beta, int order) const
{
	return DBL(order)*PhysicalConstants::eVtimesAo/_grooveSpacing/
				(cos(beta)-cos(alpha));
} // GratingDataServer::energy()

//AI: Wavelength scale. Start
double GratingDataServer::energy(double lambda) const
{
	return PhysicalConstants::eVtimesAo/lambda;
} // GratingDataServer::energy()

//AI: Wavelength scale. End


GratingDataServer::~GratingDataServer()
{
	unsigned xExt = _reflectionEfficiencies.xExtent(),
			 yExt = _reflectionEfficiencies.yExtent();
	for (unsigned xi=0; xi<xExt; ++xi)
		for (unsigned yi=0; yi<yExt; ++yi)
			delete _reflectionEfficiencies[xi][yi];
	unsigned N = _reflectionEfficienciesCorrValues.size();
	for (unsigned n=0; n<N; delete _reflectionEfficienciesCorrValues[n++]) ;
} // GratingDataServer::~GratingDataServer()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<GratingDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	GratingDataServer *null = 0;
	_collection = new vector<GratingDataServer *>(_nEntries, null);

	const XmmInstrument &rgs = XmmPayLoad::instrument(
													instrument->RGSinChain());
	/*	Get incidence angle. It can occur that the calServer is not
        correctly set up at this stage, in which case a temporary
        server needs to be instantiated. */
	CalServer *tempCalServer = new CalServer(rgs.id());
    /* AI gcc4 porting
     * 'gds' and 'mds' may be used uninitialized in this function =>
     *  must be initialized
     */
	MiscDataServer *mds = 0;
    mds = calServer.getAtom(mds);
    GeometryDataServer *gds = 0;
    gds = calServer.getAtom(gds);

    // If the calServer is not ok - use a temporary server
    if (! mds->has("INCIDENCE_ANGLE")) {
       calServer.releaseAtom(mds);
       calServer.releaseAtom(gds);
	   mds = tempCalServer->getAtom(mds);
	   gds = tempCalServer->getAtom(gds);
    }

	double alpha0 = mds->get("INCIDENCE_ANGLE") * DEG_TO_RAD +
					gds->coordFrameQuantities().gry,		//	rotation of RGA
															//	about +Y [rad]
		   grooveSpacing = mds->get("GRAT_LINE_DENS"),		// [Angstrom]
		   F = mds->get("FOCAL_LENGTH");					// [mm]
	double L = (gds->coordFrameQuantities().G -
				gds->coordFrameQuantities().M).norm();		// [mm]

	//tempCalServer->releaseAtom(mds);
	delete tempCalServer;

	CifEntry *qe = ccf->entry(&rgs, Ccf::QuantumEf),
			 *lsf = ccf->entry(&rgs, Ccf::LineSpreadFunc);
	DataSet *qeset = qe->open(),
			*lsfset = lsf->open();
	CcfConstituentTable ccftab0(lsfset, "FIGURE", 1),
						ccftab1(lsfset, "BOWS", 0, true),
						ccftab2(lsfset, "SCATTER"),
						ccftab3(lsfset, "LASCAT", 1),
						ccftab4(qeset, "RGA_EFF"),
						ccftab5(qeset, "RGA_EFFCORR", 0, true),
						ccftab6(qeset, "RGA_SELFVIGNCORR", 0, true),
						ccftab7(qeset, "RGA_INTERCEPT"),
						ccftab8(qeset, "RGA_OBSCURATE");
	(*_collection)[0] = new GratingDataServer(ccftab0,
											  ccftab1,
											  ccftab2,
											  ccftab3,
											  ccftab4,
											  ccftab5,
											  ccftab6,
											  ccftab7,
											  ccftab8,
											  alpha0,
											  grooveSpacing,
											  F/L,
											  this);

	dataSetServer->close(qeset);
	dataSetServer->close(lsfset);
} // CalAtomCollection<GratingDataServer>::CalAtomCollection()


/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
GratingDataServer *CalAtomCollection<GratingDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<GratingDataServer>::select()
