//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmGeoLinearizer.cc --- implementation of OmGeoLinearizer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2000-01-11: 1.2.0: - initial version
//		UL, 2000-07-28: 1.3.0: - bug fix in reading of POLYNOM_MAP
//		UL, 2001-05-04: 1.4.0: - access magnifier filter table as
//								 "FILTER-MAGNI" (see also SPR 2002)
//
//	AUTHOR:
//		U. Lammers (UL), Tue Jan 11 16:42:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//		[3]: CCF ICD
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/OmGeoLinearizer.cc,v 1.21 2009/07/06 14:25:19 aibarra Exp $
//

#include "OmGeoLinearizer.h"
#include "CalOalUtility.h"

using namespace std;

//
//	---( CalAtomCollection<OmGeoLinearizer> ) --------------------------------
//
/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<OmGeoLinearizer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	OmGeoLinearizer *null = 0;
	_collection = new vector<OmGeoLinearizer *>(_nEntries, null);

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::AstroMet);
	DataSet *dataSet = e->open();
	string filterName = uppercase(Xmm::toString(filter));
	if (filter == Xmm::MAGNI)
		//
		//	OM Magnifier filter is referred to as "MAGNI" in [2] + [3]
		//	but we like to continue using the long name in the SAS;
		//	since [3] is frozen this is only possible with this ugly
		//	construct
		//
		filterName.erase(5); // "MAGNIFIER" -> "MAGNI"

	CcfConstituentTable ccftab(dataSet, "FILTER-" +	filterName);
	const Table *table = ccftab.block();

	ColumnData *cdrawx = table->column("RAWX")->data(),
			   *cdrawy = table->column("RAWY")->data(),
			   *cdoffx = table->column("RAWX_OFF")->data(),
			   *cdoffy = table->column("RAWY_OFF")->data();

	int16	*rawX		= cdrawx->int16Data(),
			*rawY		= cdrawy->int16Data();
	real32	*offX		= cdoffx->real32Data(),
			*offY		= cdoffy->real32Data();

	unsigned N = table->rows();
	vector<real32> x(N), y(N), xo(N), yo(N);
	for (unsigned n=0; n<N; ++n) {
		x[n] = FLT(rawX[n]);
		y[n] = FLT(rawY[n]);
		xo[n] = offX[n];
		yo[n] = offY[n];
	}
	double plateScale = table->attribute("PLTSCALE")->asReal();

	//
	//	read polynomial coefficients
	//
	vector<double>	xa, ya, ixa, iya;
	if (dataSet->hasBlock("POLYNOM_MAP")) {
		table = dataSet->table("POLYNOM_MAP");
		ColumnData *cdfilterId = table->column("FILTER_ID")->data(),
				   *cdxav = table->column("XPOLYCOEF")->data(),
				   *cdyav = table->column("YPOLYCOEF")->data();

		//	search the right entry
		unsigned N = table->rows(), n;
	    string filterName = uppercase(Xmm::toString(filter));
		for (n=0; n<N; ++n)
			if (filterName == CalOalUtility::stringColumnRow(cdfilterId, n)) {
				unsigned M = cdxav->elements();
				real32 *xav = cdxav->real32Data(),
					   *yav = cdyav->real32Data();
				for (unsigned m=0; m<M; ++m) {
					unsigned pos = cdxav->position(n, m);
					xa.push_back(xav[pos]);
					ya.push_back(yav[pos]);
				}
				break;
			}

		if (n>=N)
			errstr<<"`"<<table->qualifiedName()<<"' does not contain an "
					"entry for filter `"<<Xmm::toString(filter)
				  <<"' - polynomial offset calculation will not work"
				  <<Sas::warning("NoSuchFilterEntry");

		delete cdfilterId;
		delete cdxav;
		delete cdyav;
	} else
		errstr<<"`"<<dataSet->qualifiedName()<<"' does not contain a table "
				"with name `POLYNOM_MAP' - polynomial offset calculation will "
				"not work"<<Sas::warning("NoSuchTable");

   /* Read in the inverse transformation coefficients */
	if (dataSet->hasBlock("POLYNOM_MAP2")) {
		table = dataSet->table("POLYNOM_MAP2");
		ColumnData *cdfilterId = table->column("FILTER_ID")->data(),
				   *cdxav = table->column("XPOLYCOEF")->data(),
				   *cdyav = table->column("YPOLYCOEF")->data();

		//	search the right entry
		unsigned N = table->rows(), n;
	    string filterName = uppercase(Xmm::toString(filter));
		for (n=0; n<N; ++n)
			if (filterName == CalOalUtility::stringColumnRow(cdfilterId, n)) {
				unsigned M = cdxav->elements();
				real32 *xav = cdxav->real32Data(),
					   *yav = cdyav->real32Data();
				for (unsigned m=0; m<M; ++m) {
					unsigned pos = cdxav->position(n, m);
					ixa.push_back(xav[pos]);
					iya.push_back(yav[pos]);
				}
				break;
			}

		if (n>=N)
			errstr<<"`"<<table->qualifiedName()<<"' does not contain an "
					"entry for filter `"<<Xmm::toString(filter)
				  <<"' - polynomial offset calculation will not work"
				  <<Sas::warning("NoSuchFilterEntry");

		delete cdfilterId;
		delete cdxav;
		delete cdyav;
	} else
		errstr<<"`"<<dataSet->qualifiedName()<<"' does not contain a table "
				"with name `POLYNOM_MAP2' - inverse polynomial offset "
                "calculation will not work"<<Sas::warning("NoSuchTable");

	//
	//	construct new OmGeoLinearizer
	//
	OmGeoLinearizer *&omGeoLinearizer = (*_collection)[0];
	omGeoLinearizer = new OmGeoLinearizer(x, y, xo, yo, xa, ya, ixa, iya,
                                         plateScale, this);

	//	tidy up
	delete cdrawx;
	delete cdrawy;
	delete cdoffx;
	delete cdoffy;

	dataSetServer->close(dataSet);
} // CalAtomCollection<OmGeoLinearizer>::CalAtomCollection()


void OmGeoLinearizer::dump(ostream &os)
{
	os<<"OmGeoLinearizer{\n";
	os<<"}";
} // OmGeoLinearizer::dump()

/* AI gcc4 porting
 * too few template-parameter-lists => template<>
 */
template<>
OmGeoLinearizer *CalAtomCollection<OmGeoLinearizer>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<OmGeoLinearizer>::select


//
//	---( OmGeoLinearizer> ) -----------------------------------------------
//
OmGeoLinearizer::OmGeoLinearizer(const vector<real32> &rawX,
			 					 const vector<real32> &rawY,
								 const vector<real32> &deltaX,
								 const vector<real32> &deltaY,
								 const vector<double>  &xa,
								 const vector<double>  &ya,
								 const vector<double>  &ixa,
								 const vector<double>  &iya,
								 double plateScale,
								CalAtomCollection<OmGeoLinearizer> *parent)
	: CalChild<CalAtomCollection<OmGeoLinearizer> >(parent), _ax(xa), _ay(ya),
      _iax(ixa), _iay(iya), _plateScale(plateScale)
{
	_xoffInter =
		new Slatec::Bivariate2DInterpolantScattered<float>(rawX, rawY, deltaX);
	_yoffInter =
		new Slatec::Bivariate2DInterpolantScattered<float>(rawX, rawY, deltaY);
} // OmGeoLinearizer::OmGeoLinearizer()


void OmGeoLinearizer::additionalPoints(unsigned nPoints)
{
	_xoffInter->additionalDataPoints(nPoints);
	_yoffInter->additionalDataPoints(nPoints);
} // OmGeoLinearizer::additionalPoints()


void OmGeoLinearizer::interpolatedDeltas(const CalReal32Vector &rawX,
										 const CalReal32Vector &rawY,
										 CalReal32Vector &deltaX,
										 CalReal32Vector &deltaY,
										 CalReal32Vector &sigmaX,
										 CalReal32Vector &sigmaY) const
{
	unsigned N = rawX.size();
	for (unsigned n=0; n<N; ++n) {
		deltaX[n] = (*_xoffInter)(FLT(rawX[n]), FLT(rawY[n]));
		deltaY[n] = (*_yoffInter)(FLT(rawX[n]), FLT(rawY[n]));
		//	we do not know uncertainties - set the sigma vectors to 0
		sigmaX[n] = 0.f;
		sigmaY[n] = 0.f;
	}
} // OmGeoLinearizer::interpolatedDeltas()


void OmGeoLinearizer::polynomialDeltas(const CalReal32Vector &rawX,
									   const CalReal32Vector &rawY,
									   CalReal32Vector &deltaX,
									   CalReal32Vector &deltaY,
									   CalReal32Vector &sigmaX,
									   CalReal32Vector &sigmaY) const
{
	if (_ax.size() != 36 || _ay.size() != 36)
		errstr<<"Vector with polynomial coefficients too small - size "
				"is "<<_ax.size()<<" - cannot evaluate 7th order "
				"polynomial"<<Sas::error("InsufficientData");

   /* Calculate the polynomial expansion */
    polycalc(rawX, rawY, deltaX, deltaY, sigmaX, sigmaY, _ax, _ay);

} // OmGeoLinearizer::polynomialDeltas()

void OmGeoLinearizer::inversepolynomialDeltas(const CalReal32Vector &rawX,
							 		   const CalReal32Vector &rawY,
									   CalReal32Vector &deltaX,
									   CalReal32Vector &deltaY,
									   CalReal32Vector &sigmaX,
									   CalReal32Vector &sigmaY) const
{
	if (_iax.size() != 36 || _iay.size() != 36)
		errstr<<"Vector with polynomial coefficients too small - size "
				"is "<<_iax.size()<<" - cannot evaluate 7th order "
				"polynomial"<<Sas::error("InsufficientData");

   /* Calculate the polynomial expansion */
    polycalc(rawX, rawY, deltaX, deltaY, sigmaX, sigmaY, _iax, _iay);

} // OmGeoLinearizer::inversepolynomialDeltas()

void OmGeoLinearizer::polycalc(const CalReal32Vector &rawX,
							   const CalReal32Vector &rawY,
								     CalReal32Vector &deltaX,
									 CalReal32Vector &deltaY,
									 CalReal32Vector &sigmaX,
									 CalReal32Vector &sigmaY,
							   const std::vector<real64> &cx,
							   const std::vector<real64> &cy) const
{
    unsigned N = rawX.size();
    for (unsigned n=0; n<N; ++n) {
        double x1 = DBL(rawX[n]) - 1024.5,
               y1 = DBL(rawY[n]) - 1024.5,
               x2 = x1*x1,
               x3 = x2*x1,
               x4 = x2*x2,
               x5 = x4*x1,
               x6 = x3*x3,
               x7 = x6*x1,
               y2 = y1*y1,
               y3 = y2*y1,
               y4 = y2*y2,
               y5 = y4*y1,
               y6 = y3*y3,
               y7 = y6*y1;

        deltaX[n] = cx[0] + cx[1]*y1 + cx[2]*y2 + cx[3]*y3 + cx[4]*y4 +
                    cx[5]*y5 + cx[6]*y6 + cx[7]*y7 +

                    cx[8]*x1 + cx[9]*x1*y1 + cx[10]*x1*y2 + cx[11]*x1*y3+
                    cx[12]*x1*y4 + cx[13]*x1*y5 + cx[14]*x1*y6 +

                    cx[15]*x2 + cx[16]*x2*y1 + cx[17]*x2*y2 +
                    cx[18]*x2*y3 + cx[19]*x2*y4 + cx[20]*x2*y5 +

                    cx[21]*x3 + cx[22]*x3*y1 + cx[23]*x3*y2 +
                    cx[24]*x3*y3 + cx[25]*x3*y4 +

                    cx[26]*x4 + cx[27]*x4*y1 + cx[28]*x4*y2 +
                    cx[29]*x4*y3 +

                    cx[30]*x5 + cx[31]*x5*y1 + cx[32]*x5*y2 +

                    cx[33]*x6 + cx[34]*x6*y1 +

                    cx[35]*x7;

        deltaY[n] = cy[0] + cy[1]*y1 + cy[2]*y2 + cy[3]*y3 + cy[4]*y4 +
                    cy[5]*y5 + cy[6]*y6 + cy[7]*y7 +
                    cy[8]*x1 + cy[9]*x1*y1 + cy[10]*x1*y2 + cy[11]*x1*y3+
                    cy[12]*x1*y4 + cy[13]*x1*y5 + cy[14]*x1*y6 +

                    cy[15]*x2 + cy[16]*x2*y1 + cy[17]*x2*y2 +
                    cy[18]*x2*y3 + cy[19]*x2*y4 + cy[20]*x2*y5 +

                    cy[21]*x3 + cy[22]*x3*y1 + cy[23]*x3*y2 +
                    cy[24]*x3*y3 + cy[25]*x3*y4 +

                    cy[26]*x4 + cy[27]*x4*y1 + cy[28]*x4*y2 +
                    cy[29]*x4*y3 +

                    cy[30]*x5 + cy[31]*x5*y1 + cy[32]*x5*y2 +

                    cy[33]*x6 + cy[34]*x6*y1 +

                    cy[35]*x7;
        sigmaX[n] = sigmaY[n] = 0.f;
    }
} // OmGeoLinearizer::polycalc


OmGeoLinearizer::~OmGeoLinearizer()
{
	delete _xoffInter;
	delete _yoffInter;
} // OmGeoLinearizer::~OmGeoLinearizer()


