//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SpecQualityServer.cc --- implementation of miscallaneous CalAtomsCommon classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 2009-04-13: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2009-04-13
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/SpecQualityServer.cc,v 1.6 2010/04/12 14:45:25 rsaxton Exp $
//

#include "SpecQualityServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//

/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
CalAtomCollection<SpecQualityServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
	_nEntries	= _nModes; // Max number of entries possible
	SpecQualityServer *null = 0;
	_collection = new vector<SpecQualityServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::SpecQual);
	DataSet *dataSet = e->open();

      // Spectral Quality
        Table *btable=0;

      // Depending on the instrument open the correct block(s)
        Table *table = dataSet->table(e->queryBlockName("EMOS1_SPQ"));
        if (instrument->id() == Xmm::EMOS2)
           table = dataSet->table(e->queryBlockName("EMOS2_SPQ"));
        else if (instrument->id() == Xmm::EPN) 
           table = dataSet->table(e->queryBlockName("EPN_SPQ"));
        else if (instrument->id() == Xmm::RGS1) {
           table = dataSet->table(e->queryBlockName("RGS1_SPQ_WAVE"));
           btable = dataSet->table(e->queryBlockName("RGS1_SPQ_BETA"));
        }
        else if (instrument->id() == Xmm::RGS2) {
           table = dataSet->table(e->queryBlockName("RGS2_SPQ_WAVE"));
           btable = dataSet->table(e->queryBlockName("RGS2_SPQ_BETA"));
        }


      // Read the arrays from this block
	ColumnData *cdmodeId = table->column("MODE_ID")->data(),
		   *cdngrp = table->column("N_GROUPS")->data(),
		   *cdflag = table->column("SPE_FLAG")->data();

	int8	*modeId = cdmodeId->int8Data();
	int8    *groups = cdngrp->int8Data();

      // The range information
        unsigned N = table->rows();
        vector<vector<short> > ranges;

        for (unsigned n=0; n<N; ++n) {
            vector<short> r0;
            CalOalUtility::fillVector(table, "SPE_FLAG", n, r0);
            ranges.push_back(r0);
        }

      /* If RGS read the conversion factors from lambda and beta to channels */
        vector<vector<short> > beta_ranges(N);
        vector<int> lambda_nchan(2), beta_nchan(2);
        //vector<double> lambda_min(2), lambda_max(2), 
          //     beta_min(2), beta_max(2);

        if ( instrument->id() == Xmm::RGS1 || instrument->id() == Xmm::RGS2) {

           lambda_nchan[0] = table->attribute("DETCHAN1")->asInt();
           beta_nchan[0] = btable->attribute("DETCHAN1")->asInt();
           lambda_nchan[1] = table->attribute("DETCHAN2")->asInt();
           beta_nchan[1] = btable->attribute("DETCHAN2")->asInt();

          /* Don't need these conversions now 
           lambda_max[0] = table->attribute("WAVEMAX1")->asReal();
           lambda_min[0] = table->attribute("WAVEMIN1")->asReal();
           beta_min[0] = btable->attribute("BETAMIN1")->asReal();
           beta_max[0] = btable->attribute("BETAMAX1")->asReal();
           lambda_max[1] = table->attribute("WAVEMAX2")->asReal();
           lambda_min[1] = table->attribute("WAVEMIN2")->asReal();
           beta_min[1] = btable->attribute("BETAMIN2")->asReal();
           beta_max[1] = btable->attribute("BETAMAX2")->asReal();
         */

          // Read in the BETA ranges - NB for this to work the MODES and 
          // GROUPS have to be the same as in the _WAVE in the _BETA
          // tables. This is far from ideal...
           for (unsigned n=0; n<btable->rows(); ++n) {
              vector<short> r0;
              CalOalUtility::fillVector(btable, "SPE_FLAG", n, r0);
              beta_ranges[n]=r0;
           }
        }

      /* Read in the Energy resolutions */
      // Depending on the instrument open the correct block
        Table *rtable = dataSet->table(e->queryBlockName("EMOS1_RESOLUTION"));
        if (instrument->id() == Xmm::EMOS2)
           rtable = dataSet->table(e->queryBlockName("EMOS2_RESOLUTION"));
        else if (instrument->id() == Xmm::EPN) 
           rtable = dataSet->table(e->queryBlockName("EPN_RESOLUTION"));
        else if (instrument->id() == Xmm::RGS1)
           rtable = dataSet->table(e->queryBlockName("RGS1_RESOLUTION"));
        else if (instrument->id() == Xmm::RGS2)
           rtable = dataSet->table(e->queryBlockName("RGS2_RESOLUTION"));

        vector<float> res_energy(rtable->rows()), 
                      resolution(rtable->rows());

       // delete cdenergy;
       // delete cdres;

       /*
        ColumnData *cdenergy = rtable->column("ENERGY")->data(),
                   *cdres = rtable->column("RESOLUTION")->data();

        cout<<"Got the col data"<<endl;
        real32   *endata = cdenergy->real32Data(),
                 *resdata = cdres->real32Data();
        cout<<"Mapped it"<<endl;

        for (unsigned n=0; n<rtable->rows(); ++n) {
            res_energy[n]=endata[n];
            resolution[n]=resdata[n];
        }
       */
        
        CalOalUtility::fillVector(rtable, "ENERGY", res_energy);
        CalOalUtility::fillVector(rtable, "RESOLUTION", resolution);

        //cout<<"RES[0]: "<<res_energy[0]<<" "<<resolution[0]<<endl;
        //cout<<"RES[2]: "<<res_energy[1]<<" "<<resolution[1]<<endl;

      /* Read in the Gain limits arrays */
        Table *gtable = dataSet->table(e->queryBlockName("GAINFIT_LIMIT"));

        ColumnData *cdgmodId = gtable->column("MODE_ID")->data(),
                   *cdinstId = gtable->column("INSTRUMENT_ID")->data(),
                   *cdsmin = gtable->column("GSLOP_MIN")->data(),
                   *cdsmax = gtable->column("GSLOP_MAX")->data(),
                   *cdomin = gtable->column("GOFFS_MIN")->data(),
                   *cdomax = gtable->column("GOFFS_MAX")->data();

	int8	*gmodeId = cdgmodId->int8Data();
        real32  *slope_min = cdsmin->real32Data(),
                *slope_max = cdsmax->real32Data(),
                *off_min = cdomin->real32Data(),
                *off_max = cdomax->real32Data();

      /* Fill in the values for each mode */
	for (unsigned n=0; n<table->rows(); ++n) {

	    if (modeId[n] > _nModes) {
		errstr<<"Illegal mode number in row #"
			  <<n<<" of CCF file "
			  <<Xmm::toString(instrument->id())<<"/"
			  <<Ccf::toString(Ccf::ModeParam)
			  <<Sas::error("IllegalValues");
            }

	    //unsigned pos = index(modeId[n]);

         // Find the gain limits entry corresponding to this mode
            short gl = -999;
            for (unsigned g=0; g<gtable->rows(); ++g)
                if ( Xmm::toInstrument(stringColumnRow(cdinstId, g)) ==
                        instrument->id() && gmodeId[g] == modeId[n])
                            gl=g;

           /* Some instruments and/or modes may not have gain limit values
              so if not found set to the first value without a warning */
            if (gl == -999) gl = 0;

         // Set the values for this mode
	    (*_collection)[n] = new SpecQualityServer;

	    (*_collection)[n]->_ngroups = groups[n];
	    (*_collection)[n]->_ranges = ranges[n];
	    (*_collection)[n]->_beta_ranges = beta_ranges[n];
	    (*_collection)[n]->_gslope_min = slope_min[gl];
	    (*_collection)[n]->_gslope_max = slope_max[gl];
	    (*_collection)[n]->_offset_min= off_min[gl];
	    (*_collection)[n]->_offset_max= off_max[gl];

	    (*_collection)[n]->_beta_nchan = beta_nchan[n];
	    (*_collection)[n]->_lambda_nchan = lambda_nchan[n];

           /*
	    (*_collection)[n]->_lambda_min = lambda_min[n];
	    (*_collection)[n]->_lambda_max = lambda_max[n];
	    (*_collection)[n]->_beta_min= beta_min[n];
	    (*_collection)[n]->_beta_max = beta_max[n];
           */
         
         // Resolution not currently mode dependent
            (*_collection)[n]->_res_energy = res_energy;
            (*_collection)[n]->_resolution = resolution;
	}


      // Delete the column pointers
        delete cdmodeId;
        delete cdngrp;
        delete cdflag;
	delete cdgmodId;
	delete cdsmin;
	delete cdsmax;
	delete cdomin;
	delete cdomax;
        //delete cdenergy;
        //delete cdres;

	dataSetServer->close(dataSet);
} // CalAtomCollection<SpecQualityServer>::CalAtomCollection


/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
SpecQualityServer *CalAtomCollection<SpecQualityServer>::select(
					Ccd,
					Node,
					unsigned modeOffset) const
{
	unsigned n = index(modeOffset);

	return (*_collection)[n];
} // CalAtomCollection<SpecQualityServer>::select
